/*
 * $Id: fparse.c,v 1.1 1998/06/15 04:18:22 elkner Exp $
 *
 * Author:  Jens Elkner            elkner@ivs.cs.uni-magdeburg.de
 * Project: Jesalfa     http://ivs.cs.uni-magdeburg.de/~elkner/webtools/jesalfa/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html or ./gpl.html
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "main.h"
#include "debug.h"
#include "tree.h"
#include "utils.h"
#include "register.h"
#include "ip_list.h"
#include "fparse.h"

static char *GetAddress(const char *);
static char *CleanLine(const char *);
static int Logtype_cmp(t_logtype *, t_logtype *);
static int CheckNewGroup(int, const char *, int);
static int RegisterLogtype(char*, int *, int *, int *, t_logtype *, int);
static void DeleteLogtype(t_logtype *);
static int GetLogtypeGrpId(const char *, int);
static int RegisterOutFile(char *);
static int RegisterExcludeProto(char *);
static void WrongOutFileKey(char *);
static int ExcludeProto (char *);

tree *logtype_root;
tree hierlog_root;
int maxlogtypegrpid;
static const char *const w_space = " \t\n\r";
t_wordlist *outfilelist = NULL;
t_wordlist *protolist = NULL;
t_hitsize udp;

int
ReadCountryCodes(const char *infile)
{
    FILE *fd;
    char *token = NULL;
    char tmp_line[BUFSIZE];
    
    if ((fd = fopen(infile, "r")) == NULL) {
	token = strerror(errno);
	fprintf(errfd,
		"Can't open the country code file %s for reading: %s\n",
		infile, token);
	exit(1);
    }
    tree_init(&tld_root);
    lineno = 0;
    while (fgets(tmp_line, BUFSIZE, fd) != NULL) {
	lineno++;
	token = CleanLine(tmp_line);
	if ( token == NULL )
	    continue;
	RegisterTLD(token, IP);
    }
    RegisterTLD("IP", IP);
    RegisterTLD("???", IP);
    close(fd);
    return 0;
 }

int
ReadrcFile(const char *infile, ip_acl **ip_list) {
    FILE *fd;
    char *token = NULL;
    char tmp_line[BUFSIZE];
    t_logtype *entry, *lt;
    int flag, idx, grpid, open_groups, tmp;
    
    flag = idx = grpid = open_groups = tmp = 0;
    if ((fd = fopen(infile, "r")) == NULL) {
	token = strerror(errno);
	fprintf(errfd,
		"Can't open the configuration file %s for reading: %s\n",
		infile, token);
	exit(1);
    }
    tree_init(&logtype_root);
    entry = (t_logtype *) malloc(sizeof(t_logtype));
    memadd(sizeof(t_logtype),"t_logtype entry ReadrcFile");
    entry->id = idx++;		
    entry->name = NULL;
    memset(tmp_line, '\0', BUFSIZE);
    lineno=0;
    while (fgets(tmp_line, BUFSIZE, fd) != NULL) {
	lineno++;
	token = CleanLine(tmp_line);
	if ( token == NULL )
	    continue;
	if ( token[strlen(token)-1] == ':' ) {
	    /* a new keyword */
	    if ( open_groups != 0 ) {
		/* a group {} is not closed with a } */
		fprintf(errfd,
			"%3d: Error: Group not closed before the start "
			"of a new section (%s)\n",lineno, token);
		entry->name = NULL;
		DeleteLogtype(entry);
		return (-1);
	    }
	    switch (flag) {
		case 1 : {
		    maxlogtypegrpid = grpid;
		    break;;
		}
		default :
		    /* nothing */;
	    }
	    idx=0;
	    grpid=0;
	    if ( strcasecmp(token,"logtypes:") == 0 )
		flag = 1;
	    else if ( strcasecmp(token,"files:") == 0 )
		flag = 2;
	    else if ( strcasecmp(token,"exclude-clients:") == 0 )
		flag = 3;
	    else if ( strcasecmp(token,"exclude-protos:") == 0 )
		flag = 4;
	    else
		flag = 0;
	    continue;
	}
	switch (flag) {
	    case 1 : {
		tmp = RegisterLogtype(token,&idx,&grpid,&open_groups,entry,lineno);
		if (tmp < 0) {
		    entry->name = NULL;
		    DeleteLogtype(entry);
		    return(-1);
		}
		continue;
	    }
	    case 2 : {
		if ( RegisterOutFile(token) )
		    return -1;
		continue;
	    }
	    case 3 : {
		addToIPACL(ip_list,token,lineno);
		continue;
	    }
	    case 4 : {
		RegisterExcludeProto(token);
		continue;
	    }
	    default :
		/* ignore the line */;
	}
    }
    close(fd);
    entry->name = 0;
    DeleteLogtype(entry);
    return (0);
}

static int
RegisterOutFile(char *line)
{
    char *t;
    FILE *fd;
    t_wordlist *tmp;
    char buffer[1024];
    
    t = strchr(line,'=');
    if (!t) {
	WrongOutFileKey(line);	
	return -1;
    }
    *t = '\0';
    if (outpath) {
	strcpy(buffer,outpath);
	strcat(buffer,"/");
	strcat(buffer,line);
    }
    else
	strcpy(buffer,line);
    if ((fd = fopen(buffer, "w")) == NULL) {
	line = strerror(errno);
	fprintf(errfd,"Can't open the output file %s for writing:\n%s\n",buffer,line);
	*t = '=';
	return -1;
    }
    close(fd);
    *t = '=';
    t++;
    if ( (strlen(t)) < (strlen("t-s-p-N")) ) {
	WrongOutFileKey(line);	
	return -1;
    }	
    if ( (*t != 't') && (*t != 'h') ) {
	*(t+1) = '\0';
	fprintf(errfd,"%3d: Error: Type not supported (%s)\n",
		lineno,t);
	return -1;
    }
    t += 2;
    switch (*t) {
	case 't':
	case 's':
	    do_domain++;
	    break;
	case 'C':
	case 'S':
	    do_content++;
	    break;
	case 'e':
	    do_extension++;
	    break;
	case 'p':
	    do_proto++;
	    break;
	default:
	    *(t+1) = '\0';
	    fprintf(errfd,"%3d: Error: StatType not supported (%s)\n", lineno,t);
	    return -1;
    }
    t += 2;
    if (  (*t != 'p') && (*t != 'n') ) {
	*(t+1) = '\0';
	fprintf(errfd,"%3d: Error: Presentation Format not supported (%s)\n",
		lineno,t);
	return -1;
    }
    t += 2;
    if (isdigit(*t) )
	debug(1,"file: section ok.");
    else {
	fprintf(errfd,"%3d: Error: %s is not a number\n",
		lineno,t);
	return -1;
    }
    if (!outfilelist) {
	outfilelist = (t_wordlist *) malloc(sizeof(t_wordlist));
	memadd(sizeof(t_wordlist),"Outfile entry");
	outfilelist->key = savestr(line);
	outfilelist->next = NULL;
	return 0;
    }
    tmp = outfilelist;
    while (tmp->next)
	    tmp = tmp->next;
    tmp->next = (t_wordlist *) malloc(sizeof(t_wordlist));
    memadd(sizeof(t_wordlist),"Outfile entry");
    tmp = tmp->next;
    tmp->key = savestr(line);
    tmp->next = NULL;
    return 0;
}

static int
RegisterExcludeProto(char *proto) 
{
    t_wordlist *tmp;

    if (!protolist) {
	protolist = (t_wordlist *) malloc(sizeof(t_wordlist));
	memadd(sizeof(t_wordlist),"Exclude Protocol entry");
	protolist->key = savestr(proto);
	protolist->next = NULL;
	return 0;
    }
    tmp = protolist;
    while (tmp->next)
	tmp = tmp->next;
    tmp->next = (t_wordlist *) malloc(sizeof(t_wordlist));
    memadd(sizeof(t_wordlist),"Exclude Protocol entry");
    tmp = tmp->next;
    tmp->key = savestr(proto);
    tmp->next = NULL;
    return 0;   
}

static void
WrongOutFileKey(char *line)
{
    fprintf(errfd,"%3d: Error: Key Format not supported (%s)\n",
	    lineno,line);
}

static int
RegisterLogtype(char *token, int *idx, int *grpid, int *open_groups,
		t_logtype *entry, int line)
{
    t_logtype *logtype, *lt;
    const int flag = 1;
    int tmp;
    
    tmp = CheckNewGroup(*open_groups, token, line);
    *open_groups += tmp;
    switch (tmp) {
	case 9999 :
	    return -1;
	case -1 :
	    return 0;
	case 0 : {
	    if (*open_groups == 0)
		(*grpid)++;
	    (*idx)++;
	    break;
	}
	case 1 : { 
	    (*grpid)++;
	    return 0;
	}
    }
    entry->name = token;
    /* falls noch nicht registriert */
    logtype = tree_srch(&logtype_root,Logtype_cmp,entry);
    if (! logtype) {
	logtype= (t_logtype *) malloc(sizeof(t_logtype));
	memadd(sizeof(t_logtype),"t_logtype RegisterLogtype");
	logtype->id = *grpid - 1;
	logtype->name = savestr(token);
	lt = tree_add(&logtype_root,Logtype_cmp,logtype,NULL);
	if (! lt) {
	    fprintf(errfd,"line %3d: Error adding node for %d = %s\n",
		    line, *idx , logtype->name);
	    entry->name = NULL;
	    DeleteLogtype(logtype);
	    return -2;
	}
	debug(1,"line %3d: %2d. LOGTYPE %s added (group %d)",
	      line,*idx, logtype->name, *grpid);
    }
    else {
	idx--; grpid--;
	fprintf(errfd,
		"line %3d: Logtype %s already in group %d. Ignored!!!\n",
		line, *idx, logtype->id, logtype->name);
    }
    return 0;
}

static int
Logtype_cmp(t_logtype *l,t_logtype *r)
{
    return strcasecmp(l->name,r->name);
}

static char *
CleanLine(const char *line) {
    char *token = NULL;

    token = (char *) line;
    token[strlen(line)-1] = ' ';
    /* skip leading and trailing white space */
    token += strspn(line, w_space);
    if ( *token == '#' ) {
	return (NULL);
    }
    token[strcspn(token, w_space)] = '\0';
    if ( *token == '\0' ) {
	return (NULL);
    }
    return (token);
}

static int
CheckNewGroup(int g, const char *t, int line)
{
    switch (*t) {
	case '{' : {
	    if ( g > 0 ) {
		fprintf(errfd, "%3d: Error: Nested Groups not allowed.",
			line);
		return(9999);
	    }
	    else
		return(1);
	}
	case '}' : {
	    if ( g < 1 ) {
		fprintf(errfd, "%3d: Error: Unexpected end of group.",
			line);
		return(9999);
	    }
	    else
		return(-1);
	}
	default :
	    return 0;
    }
}

int
CleanLogtypes(tree **t)
{
    tree_mung(t, DeleteLogtype);
    return 0;
}

static void
DeleteLogtype(t_logtype *lt)
{
  if (!lt) return;
  delstr(lt->name);
  memdel(sizeof(t_logtype),"t_logtype");
  free(lt);
}

int
ParseFile(const char *infile, ip_acl *ip_list)
{
  FILE *fd;
  char *token, *token2, *proto, *wpurl;
  static char tmp_line[BUFSIZE];
  int grpid = 0;
  int invalid_url;
  
  /* LOGFILE Line: "Timestamp Elapsed Client Action/Code Size Method
     URI Ident Hierarchy/From Content" */
 
  int time_sec, time_usec, elapsed;
  char client[16];		/* IP-Address like 111.111.111.111 */
  char action[32];
  int code, size;
  char method[16];
  char uri[URLBUF];
  char ident[64];
  char hierarchy[32];
  char from[256];
  char content[128];
  struct in_addr address;
  
  if (strcmp(infile, "-") == 0) fd = stdin;
  else {
    if ((fd = fopen(infile, "r")) == NULL) {
	token = strerror(errno);
	fprintf(errfd, "Can't open access log file %s: %s\n",
		infile, token);
      return(-1);
    }
  }
  memset(tmp_line, '\0', BUFSIZE);
  lineno = 0;
  udp.no = 0; udp.bytes = 0;
  while (fgets(tmp_line, BUFSIZE, fd) != NULL) {
    lineno++;
    if ( (token = strchr(tmp_line, '\n'))) {
      *token = '\0';
    }    
    if (tmp_line[0] == '\0')
      continue;
    code = sscanf(tmp_line,"%d.%d %d %s %s %d %s %s %s %s %s",
	   &time_sec, &time_usec, &elapsed, client,
	   action, &size, method, uri, ident, hierarchy, content);
    if ( code != 11 ) {
	if (verbose)
	    fprintf(errfd, "Error: %7d  Invalid access.log entry ignored:\n"
		    "%s\n", lineno, tmp_line);
	continue;
    }
    /* check whether to exclude this client */
    if (ip_list) {
	if ( (address.s_addr = inet_addr(client)) == -1) {
	    if (verbose)
		fprintf(errfd, "ERROR: %7d  Invalid client IP address - "
			"the following entry is ignored:\n%s\n",
			lineno, tmp_line);
	    continue;
	}
	if ( ip_access_check(address, ip_list) == IP_DENY )
	    continue;
    }
    /* extract the protocol and strip it off from the uri */ 
    token = strstr(uri,"://");
    if (token) {
	*token = '\0';
	proto = uri;
	wpurl = token + 3;
    }
    else {
	proto = NULL;
	wpurl = uri;
	if (verbose)
	    fprintf(errfd, "Error: %7d  no proto for URI %s\n", lineno, wpurl);
    }
    if ( protolist && proto ) {
	/* i.e. accesslog lines with no proto in it are always logged */
	if ( ExcludeProto(proto) )
	    continue;
    }
    /* separate action from code */
    token = strchr(action, '/');
    if (! token) {
	if (verbose)
	    fprintf(errfd, "Error: %7d  Invalid access.log entry ignored:\n"
		    "%s\n", lineno, tmp_line);
	continue;
    }
    *token = '\0';
    invalid_url = 0;
    if ( (strcasecmp(action,"ERR_INVALID_URL") == 0) ||
	 (strcasecmp(action,"ERR_DNS_FAIL") == 0) ) {
	invalid_url++;
	debug(2, "invalid URL - counted to TLD.SLD ???.???, "
	      "Extensions are ignored");
    }
    grpid = GetLogtypeGrpId(action, size);

    /* do the registration stuff */
    if ( do_extension && (! invalid_url) ) {
	token = strchr(wpurl,'?'); /* next 3 lines cut off parameters but */
	if (token)		   /* preserve the '?'- newer squids do not */
	    *(++token) = '\0';	   /* log them anymore :( ... */
	token = strrchr(wpurl,'/');
	if (token)
	    token = strrchr(token,'.');
	if (token && ( *(++token) != '\0') ) {
	    RegisterExtension(token,grpid,size);
	}
	else
	    RegisterExtension("-",grpid,size);
    }
    if (do_domain) { 
	if ( invalid_url )
	    RegisterServer(NULL,grpid,size,0);
	else {
	    token = GetAddress(wpurl);
	    RegisterServer(token,grpid,size,1);
	}
    }
    if (do_proto && proto) {
	    RegisterProto(proto,grpid,size);
    }
    if ( do_content && (*content != '-') ) {
	RegisterContent(content,grpid,size);
    }
    debug(2,"%9d: %s %s %d %s %s %s",
	  lineno, client, action, size, proto ? proto : "", wpurl, content);
  }
  fclose(fd);
  return 0;
}

static int
ExcludeProto (char *proto) 
{
    int exclude = 0;
    t_wordlist *p, *n;

    if (proto)
	for (p = protolist; p; p = p->next) {
	    if ( strcmp(proto,p->key) == 0 ) {
		debug(3,"Excluding proto %s (line %d)", proto, lineno);
		exclude++;
		break;
	    }
	}
    return exclude;
}

static int
GetLogtypeGrpId(const char *t, int s)
{
    static t_logtype *lt, e;
    static char none[] = "NONE";
    
    e.name = (char *) t;
    lt = tree_srch(&logtype_root, Logtype_cmp, &e);
    if (!lt) {
	e.name = none;
	lt = tree_srch(&logtype_root, Logtype_cmp, &e);
    }
    if ( strncasecmp("UDP_",t,4) == 0 ) {
	udp.no++;
	udp.bytes += s;
    }
    return lt->id;
}

/* extract the address from the URL */
static char *
GetAddress(const char *url)
{
   char *idx = NULL;
   char *idx1 = NULL;
   int k;

   /* 'proto://' has been already removed */
   idx = (char *) url;
   /* remove document_root.* */
   if ((idx1 = strchr(idx,'/')) != NULL)
       *idx1 = '\0';
   /* remove passwd info */
   if ((idx1 = strchr(idx,'@')) != NULL)
       idx = idx1 + 1;
   /* remove port info */
   if ( (idx1 = strchr(idx,':')) != NULL)
       *idx1 = '\0';
   /* remove trailing dot */
   k = strlen(idx) - 1;
   if ( idx[k] == '.' )
       idx[k] = '\0';
   return idx;
}







