/*
 * $Id: register.c,v 1.1 1998/06/15 04:16:58 elkner Exp $
 *
 * Author:  Jens Elkner            elkner@ivs.cs.uni-magdeburg.de
 * Project: Jesalfa     http://ivs.cs.uni-magdeburg.de/~elkner/webtools/jesalfa/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html or ./gpl.html
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "ctype.h"

#include "main.h"
#include "debug.h"
#include "utils.h"
#include "tree.h"
#include "register.h"

static t_tld *NewTLD(const char *, stat_type);
static void DeleteTLD(t_tld *);
static int TLD_cmp(t_tld *, t_tld *);
static int SLD_cmp(t_sld *, t_sld *);
static void DeleteTLDs(t_tld *);
static int GetEmptyTLDs(t_tld *);
static int CheckInetAddress(const char*);

static int stackidx = 0;
static char **tld_name_stack;

static t_tld tld_entry;
static t_sld sld_entry;

tree *tld_root;
int maxtld = 0;
int maxtldlen = 0;
int maxsld = 0;
int maxsldlen = 0;

tree *ctld_root;
int maxctld = 0;
int maxctldlen = 0;
int maxcsld = 0;
int maxcsldlen = 0;

tree *xsld_root;
int maxxsld = 0;
int maxxsldlen = 0;

tree *psld_root;
int maxpsld = 0;
int maxpsldlen = 0;

int lineno = 0;

int
CheckInetAddr(const char *addr)
{
    /* check string syntax - RFC 1123 p.13 */
    
    int a1 = 0, a2 = 0, a3 = 0;
    static char addrbuf[32];
    int a4 = 0;

    if (sscanf(addr, "%d.%d.%d.%d", &a1, &a2, &a3, &a4) == 4) {
	if (a1 < 0 || a1 > 255)
	    return 0;
	if (a2 < 0 || a2 > 255)
	    return 0;
	if (a3 < 0 || a3 > 255)
	    return 0;
	if (a4 < 0 || a4 > 255)
	    return 0;
	return 1;			/* valid numeric IP address */
    }
    a2 = strlen(addr) - 1;
    a3 = 1;
    for (a1=0; a1 < a2; a1++) {
	if ( isalnum(addr[a1]) ) {
	    a3 = 0;
	    continue;
	}
	else {
	    switch (addr[a1]) {
		case '-':
		    {
			if ( a3 )
			    return 0;
			continue;
		    }
		case '.':
		    {
			if ( a3 )
			    return 0;
			a3++;
			continue;
		    }
		default:
		    return 0;
	    }
	}
    }
    return 2;
}

void
DeleteEmptyTLDs(void)
{
    int idx;

    tld_name_stack = (char **) calloc(maxtld,sizeof(char *));
    memadd(maxtld * sizeof(char *), "Temporary TLD name pointer array");
    idx = 0;
    tree_trav(&tld_root,GetEmptyTLDs);
    for (idx = 0; idx < stackidx; idx++) {
	tld_entry.name = tld_name_stack[idx];
	debug(3,"Delete empty TLD %s\n",tld_entry.name);
	tree_delete(&tld_root,TLD_cmp, &tld_entry, DeleteTLD);
    }
    memdel(maxtld * sizeof(char *), "Temporary TLD name pointer array");
    maxtld -= stackidx;
    free(tld_name_stack);
}

int
GetEmptyTLDs(t_tld *tld)
{
    if (tld->sld_root)
	return TRUE;
    tld_name_stack[stackidx] = tld->name;
    stackidx++;
    return TRUE;
}

int
RegisterTLD(const char *t, stat_type s)
{
    t_tld *tld, *tld1;
    tree **rtld_root;
    
    tld_entry.name = (char *) t;
    if ( s == IP )
	rtld_root = &tld_root;
    else
	rtld_root = &ctld_root;
    tld = tree_srch(rtld_root, TLD_cmp, &tld_entry);
    if (!tld) {
	tld = NewTLD(t,s);
	if (!tld)
    	    return -1;
	tld1 = tree_add(rtld_root, TLD_cmp, tld, NULL);
	if (!tld1) {
	    fprintf(errfd,"Error during registeration of %s TLD %s\n",
		    s == IP ? "" : "content", t);
	    return -1;
	}
    }
    else
	/* we can do this as well for content TLD because existens should be
	   have been checked before calling RegisterTLD */
	fprintf(errfd," %s TLD %s is already registered\n",
		s == IP ? "" : "content",t);
    return 0;
}

int
RegisterServer(const char *server, int id, int size, int valid_dns)
{
    char *token, *token2;
    t_tld *tld;
    t_sld *sld, *sld1;
    char unknown[] = "IP";
    char invalid[] = "???";
    short tld_invalid = 0;
    t_hitsize *hit;
    int addrtype;
    
    /* Top Level Domain */
    if (server)
	token = strrchr(server,'.');
    else
	token = NULL;
    if (!token || !valid_dns) {
	tld_entry.name = invalid;
	sld_entry.name = invalid;
    }
    else {
	/* need to check, whether it is a valid domain name wrt. RFC 1035
	   to reduce countings for Non-existent domains */
	addrtype = CheckInetAddr(server);
	if ( addrtype == 0 ) {
	    if (verbose)
		fprintf(errfd,
			"Error: %7d %s is not a valid Domain - counted to %s.%s\n",
			lineno, server, invalid,invalid);
	    tld_entry.name = invalid;
	    sld_entry.name = invalid;
	}
	else if ( addrtype == 1 ) {
	    tld_entry.name = unknown;
	    sld_entry.name = unknown;
	}
	else {
	    tld_entry.name = token + 1;
	    /* Second Level Domain */
	    *token = '\0';
	    token2 = strrchr(server,'.');
	    if ( token2 )
		sld_entry.name = token2 + 1;
	    else
		sld_entry.name = (char *) server;
	}
    }
    tld = tree_srch(&tld_root, TLD_cmp, &tld_entry);
    if (!tld) {
	if (verbose)
	    fprintf(errfd,
		    "Error: %7d %s is not a valid TLD - counted to TLD %s\n",
		    lineno, tld_entry.name, invalid);
	tld_entry.name = invalid;
 	tld = tree_srch(&tld_root, TLD_cmp, &tld_entry);
	sld_entry.name = invalid;
    }
    sld = tree_srch(&(tld->sld_root), SLD_cmp, &sld_entry);
    if (!sld) {
	sld = NewSLD(sld_entry.name, IP);
	if (!sld)
	    return -1;
	sld1 = tree_add(&(tld->sld_root), SLD_cmp, sld, NULL);
	if (!sld1) {
	    fprintf(errfd,"Error during registration of SLD %s\n",
		    sld_entry.name);
	    return -1;
	}
	sld = sld1;
	sld->tld_name = tld->name;
	debug(3,"SLD %6d \"%s.%s\" %s registered.", maxsld,
	      sld->name, tld->name, sld ? "" : " NOT");
	sld->lineno = maxsld;
	maxsld++;
    }
    hit = sld->hitsize;
    hit[id].no++;
    hit[id].bytes +=size;
    if (token)
	*token = '.';
    return 0;
}

int
RegisterContent(const char *content, int id, int size) {
    char *token, *token2;
    t_tld *tld;
    t_sld *sld, *sld1;
    short tld_invalid = 0;
    t_hitsize *hit;
    
    /* Top Level Domain */
    token = strrchr(content,'/');
    if (!token)
	return -1;
    *token = '\0';
    tld_entry.name = (char *) content;
    /* Second Level Domain */
    sld_entry.name = token + 1;
    tld = tree_srch(&ctld_root, TLD_cmp, &tld_entry);
    if (!tld) {
	if ( RegisterTLD(tld_entry.name, CONTENT) < 0 )
	    return -1;
	else
	   tld = tree_srch(&ctld_root, TLD_cmp, &tld_entry); 
    }
    sld = tree_srch(&(tld->sld_root), SLD_cmp, &sld_entry);
    if (!sld) {
	sld = NewSLD(sld_entry.name, CONTENT);
	if (!sld)
	    return -1;
	sld1 = tree_add(&(tld->sld_root), SLD_cmp, sld, NULL);
	if (!sld1) {
	    fprintf(errfd,"Error during registration of SLD %s\n",
		    sld_entry.name);
	    return -1;
	}
	sld = sld1;
	sld->tld_name = tld->name;
	debug(3,"content SLD %6d \"%s.%s\" %s registered.", maxsld,
	      sld->name, tld->name, sld ? "" : " NOT");
	sld->lineno = maxcsld;   /* well may be this shoud be called id ;-) */
	maxcsld++;
    }
    hit = sld->hitsize;
    hit[id].no++;
    hit[id].bytes +=size;
    return 0;
}

int
RegisterExtension(const char *ext, int id, int size) {
    t_sld *sld, *sld1;
    t_hitsize *hit;
    
    sld_entry.name = (char *) ext;
    sld = tree_srch(&xsld_root, SLD_cmp, &sld_entry);
    if (!sld) {
	sld = NewSLD(sld_entry.name, EXT);
	if (!sld)
	    return -1;
	sld1 = tree_add(&xsld_root, SLD_cmp, sld, NULL);
	if (!sld1) {
	    fprintf(errfd,"Error during registration of xSLD %s\n",
		    sld_entry.name);
	    return -1;
	}
	sld = sld1;
	sld->tld_name = NULL;
	debug(3,"content xSLD %6d \"%s\" %s registered.", maxxsld,
	      sld->name, sld ? "" : " NOT");
	sld->lineno = maxxsld;   /* well may be this shoud be called id ;-) */
	maxxsld++;
    }
    hit = sld->hitsize;
    hit[id].no++;
    hit[id].bytes +=size;
    return 0;
}

int
RegisterProto(const char *proto, int id, int size) {
    t_sld *sld, *sld1;
    t_hitsize *hit;
    
    sld_entry.name = (char *) proto;
    sld = tree_srch(&psld_root, SLD_cmp, &sld_entry);
    if (!sld) {
	sld = NewSLD(sld_entry.name, PROTO);
	if (!sld)
	    return -1;
	sld1 = tree_add(&psld_root, SLD_cmp, sld, NULL);
	if (!sld1) {
	    fprintf(errfd,"Error during registration of pSLD %s\n",
		    sld_entry.name);
	    return -1;
	}
	sld = sld1;
	sld->tld_name = NULL;
	debug(3,"content pSLD %6d \"%s\" %s registered.", maxpsld,
	      sld->name, sld ? "" : " NOT");
	sld->lineno = maxpsld;   /* well may be this shoud be called id ;-) */
	maxpsld++;
    }
    hit = sld->hitsize;
    hit[id].no++;
    hit[id].bytes +=size;
    return 0;
}

int
CleanTLDs(tree **t)
{
    if (!t)
	return 0;
    tree_mung(t,DeleteTLDs);
    return 0;
}

static void
DeleteTLDs(t_tld *tld)
{
    if (!tld)
	return;
    CleanSLDs(&(tld->sld_root));
    DeleteTLD(tld);
}

int
CleanSLDs(tree **t)
{
    if (!t)
	return;
    tree_mung(t, DeleteSLD);
}

static t_tld *
NewTLD(const char *name, stat_type s)
{
    t_tld *d;
    char *tld_name;
    int len;

    d = (t_tld *) malloc(sizeof(t_tld));

    if ( d == NULL ) {
	perror("Problems allocating memory for a TLD:");
	debug(3,"Problems allocating memory for %s TLD \"%s\"",
	      s == IP ? "" : "content", d->name);
    }
    else {
	d->name = savestr(name);
	len = strlen(name);
	if ( s == IP ) 
	    if (len > maxtldlen)
		maxtldlen = len;
	else
	    if ( len > maxctldlen)
		maxctldlen = len;
	tree_init(&(d->sld_root));
	memadd(sizeof(t_tld),"t_tld");
	if ( s == IP)
	    maxtld++;
	else
	    maxctld++;
    }
    debug(3,"%3d. %s TLD \"%s\"%s registered.", s == IP ? maxtld : maxctld,
	  s == IP ? "" : "content", d->name, d ? "" : " NOT");
    return d;
}

static void
DeleteTLD(t_tld *domain)
{
    if (!domain)
	return;
    debug(3, "TLD %s unregistered.",domain->name);
    delstr(domain->name);
    free(domain);
    memdel(sizeof(t_tld),"t_tld");
}

t_sld *
NewSLD(const char *name, stat_type s)
{
    t_sld *d;
    char *sld_name;
    int len;

    d = (t_sld *) malloc(sizeof(t_sld));

    if ( d == NULL ) {
	perror("Problems allocating memory for a SLD:");
	debug(3,"Problems allocating memory for SLD \"%s\"",d->name);
    }
    else {
	d->name = savestr(name);
	d->hitsize = (t_hitsize *) calloc((maxlogtypegrpid+1),
					  sizeof(t_hitsize));
	/* well, we rely on 0 ing allocated memory by calloc */
	if (!d->hitsize) {
	    perror("Problems allocating memory for a SLD:");
	    debug(3,"Problems allocating memory for SLD \"%s\"",d->name);
	    delstr(d->name);
	    free(d);
	    return(NULL);
	}
	
	len = strlen(name);
	if ( s == IP )
	    if (len > maxsldlen)
		maxsldlen = len;
	else if ( s == CONTENT )
	    if (len > maxcsldlen)
		maxcsldlen = len;
	else if ( s == EXT )
	    if (len > maxxsldlen)
		maxxsldlen = len;
        else /* s == PROTO */
	    if (len > maxpsldlen)
		maxpsldlen = len;
	memadd(sizeof(t_sld),"t_sld");
	memadd((maxlogtypegrpid+1) *sizeof(t_hitsize),"n * t_hitsize");
    }
    return d;
}

void
DeleteSLD(t_sld *sld)
{
    int i;
    t_hitsize *hit;

    if (!sld)
	return;
    debug(3,"%s.%s unregistered.", sld->tld_name ? sld->tld_name : "",
	  sld->name);
    delstr(sld->name);
    memdel((maxlogtypegrpid+1) * sizeof(t_hitsize),"n *t_hitsize" );
    free(sld->hitsize);
    free(sld);
    memdel(sizeof(t_sld),"t_sld");
}

static int
TLD_cmp(t_tld *l, t_tld *r)
{
    return strcasecmp(l->name,r->name);
}

static int
SLD_cmp(t_sld *l, t_sld *r)
{
    return strcasecmp(l->name,r->name);
}
